//*******************************************************************
//        **** MAIN routine for microSD Data-Logger ****
//*******************************************************************
//Controller		: ATmega32 (Clock: 16 Mhz)
//Compiler			: AtmelStudio 7
//Project Version	: 2.1
//Author			: CC Dharmani, Pune (India)
//					  www.dharmanitech.com 		  
//Date				: June 2018, Updated: Sep 2021
//*******************************************************************

#ifndef F_CPU
#define F_CPU 16000000UL		//Clock freq 16MHz
#endif

#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "main.h"
#include "UART_routines.h"
#include "ADC_routines.h"
#include "RTC_routines.h"
#include "ds1621.h"
#include "LCD_4bit.h"

#ifdef SD_LOGGING_ENABLED
#include "SPI_routines.h"
#include "FAT32.h"
#endif

char line1[17], line2[17];
uint8_t screen, loggingON;
float temperature;

void portInit(void)
{
  PORTA = 0x00;
  DDRA  = 0x00;
  PORTB = 0x00; 
  DDRB  = 0xB8; //MISO i/p; MOSI, SCK, SS, CD o/p
  PORTC = 0x00; 
  DDRC  = 0xFC; //LCD port o/p; i2c pins left as i/p
  PORTD = 0x38;	//pull-ups for push-button
  DDRD  = 0xC2;	//LED-G,LED-R, TXD o/p; rest i/p (includes Pushbuttons at PC5,PC4,PC3)
}

//TIMER0 initialize - prescaler:64
//desired value: 1mSec
//actual value: 1mSec (@16MHz F_CPU)
void timer0_init(void)
{
	TCCR0 = 0x00; 
	TCNT0 = 0x06; 
	TCCR0 = 0x03; 
}

volatile uint32_t tick_1ms = 0;	//1ms tick variable using timer0

//-- TIMER0 overflow Interrupt Service Routine --
ISR(TIMER0_OVF_vect)
{
	TCNT0 = 0x06;
	tick_1ms++;
}

//call this routine to initialize all peripherals
void initDevices(void)
{
	portInit();
	timer0_init();
	spiInit();
	i2cInit();
	uartInit();
	adcInit();
	LCD_init();
	ds1621_init();

	MCUCR = 0x00;
	GICR  = 0x00;
	TIMSK = 0x01;	//timer0 overflow interrupt
	sei(); 			//Enable interrupts
}

//*************************** MAIN *******************************//
//set stream pointer for using printf, scanf
FILE uart_str = FDEV_SETUP_STREAM(transmitByte, receiveByte, _FDEV_SETUP_RW);

int main(void)
{
	uint8_t fileNo, prevDate, prevMonth;
	char fileName[13], fileNameTemp[13];
	uint16_t lineNo;
	uint32_t time1=0, time2=0, time3=0;
	float voltage[8];
	
	stdin = stdout = &uart_str;  //redirecting stdio to uart

	_delay_ms(100);  //delay for VCC stabilization

	initDevices();

	GREEN_LED_ON;  //turn on green LED to indicate power on
	RED_LED_OFF; 	 //keep red LED off for now
	
	
	printf_P (PSTR("\n\n microSD Datalogger - Ver2.0"));
	printf_P (PSTR("\n*****************************\n"));
	
	LCD_DisplayString_F(1,1,PSTR("DataLogger-v2.0"));
	
	line1[16] = '\0';
	line2[16] = '\0';
	screen = 0;
	
	//-----------------  RTC setting  --------------------------
	if(SET_PRESSED)
	{
		_delay_ms(20);
		if(SET_PRESSED)
		{
			LCD_DisplayString_F(1, 1, PSTR("RTC Setting..   "));
			RTC_read();
			while(screen != 6)
			{
				if(ANY_BUTTON_PRESSED)
				{
					_delay_ms(20);		//key debouncing delay
					if(ANY_BUTTON_PRESSED) buttonResponseRTC();
					while(ANY_BUTTON_PRESSED);
					LCD_DisplayString (2, 1, line2);
				}
				_delay_ms(20);	//loop delay of 20msec
			}
		}
		
		RTC_writeTime();
		RTC_writeDate();
		screen = 0;
		_delay_ms(40);
	}
	
	//----------- SD card initialization -------------	
#ifdef SD_LOGGING_ENABLED	
	
	cardType = 0;
	uint8_t error;

	for (int i=0; i<10; i++)
	{
		error = SD_init();
		if(!error) break;
	}

	if(error)
	{
		if(error == 1) printf_P(PSTR("SD card not detected.."));
		if(error == 2) printf_P(PSTR("Card Initialization failed.."));
		LCD_DisplayString_F(2, 1, PSTR(" SD Card Error  "));
		
		blinkRedLED();
	}

	switch (cardType)
	{
		case 1:printf_P(PSTR("Standard Capacity Card (Ver 1.x) Detected!"));
		break;
		case 2:printf_P(PSTR("High Capacity Card Detected!"));
		break;
		case 3:printf_P(PSTR("Standard Capacity Card (Ver 2.x) Detected!"));
		break;
		default:printf_P(PSTR("Unknown SD Card Detected!"));
		break;
	}

	error = getBootSectorData (); //read boot sector and keep necessary data in global variables
	if(error)
	{
		printf_P (PSTR("\n\rFAT32 not found!"));  //FAT32 incompatible drive
		LCD_DisplayString_F(2, 1, PSTR(" SD Card Error  "));
		blinkRedLED();
	}

	SPI_HIGH_SPEED;	//SCK - 4 MHz
	_delay_ms(1);   	//some delay for settling new spi speed
	
#endif
	//----------- SD Card initialized -----------
	
	printf_P(PSTR("\n\nPress SET button to start datalogging..\n"));
	
	loggingON = 0;
	fileName[12] = '\0';
	fileNo = 0;
	lineNo = 1;
	prevDate = DATE;
	prevMonth = MONTH;
	
	temperature = getTemperature();	//dummy temperature read, to start temp. conversion
	_delay_ms(1000);	//temperature registers need this delay to update the latest value
	
	RESET_TIMER;
	
	while(1)
	{
		if(SET_PRESSED)
		{
			_delay_ms(40);		//key debouncing delay
			if(SET_PRESSED) 
			{
				if(loggingON) 
				{	
					loggingON = 0; 
					RED_LED_OFF; 
					printf_P(PSTR("\nDatalogging stopped.\n"));
				}
				else 
				{ 
					lineNo = 1;
					RED_LED_ON;
#ifdef SD_LOGGING_ENABLED
					if((prevDate != DATE) || (prevMonth != MONTH))	//to reset the file number to 0 if date changes
					{
						fileNo = 0;
						prevDate = DATE;
						prevMonth = MONTH;
					}
					
					//Look for filename which doesn't already exist in the SD card
					int i;
					for(i=fileNo; i<100; i++){
						sprintf(fileName,"%02x%02x%02x%02d.CSV", DATE, MONTH, YEAR, i);
						if(readFile (VERIFY, (uint8_t*)fileName) != 1) break;
					}
					if(i == 100) {
						printf_P(PSTR("\n\nFile Capacity (100 Nos.) full for today, delete files from SD card\n"));
						blinkRedLED();
					}
					else
						fileNo = i;
					
					sprintf(fileName,"%02x%02x%02x%02d.CSV", DATE, MONTH, YEAR, fileNo++);
					printf("\nFile Name: %s\n", fileName);	
					if (fileNo > 99) fileNo = 0;
					sprintf((char*)dataString, "Sr.No,Time,Temp.(C),Channel-1,Channel-2,Channel-3,Channel-4,Channel-5,Channel-6,Channel-7,Channel-8\n");
					for(int i=0; i<13; i++) fileNameTemp[i] = fileName[i];
					error = writeFile((uint8_t*)fileNameTemp);
					if(error) 
					{
						LCD_DisplayString_F(2, 1, PSTR(" SD Card Error  "));
						printf_P(PSTR("\n\n Error in writing to File.."));
						blinkRedLED();
					}
#endif
					loggingON = 1;
					printf_P(PSTR("\n\nData-logging started..\n"));
				}
			}
			while(SET_PRESSED);
			_delay_ms(40);
			time2 = tick_1ms;
		}
		
		if(loggingON)
		{
			for(int i=0; i<8; i++) voltage[i] = adcReadVolt(i);
			sprintf((char*)dataString, "%05d,%02x:%02x:%02x,%2.1f,%2.2f,%2.2f,%2.2f,%2.2f,%2.2f,%2.2f,%2.2f,%2.2f\n",
				    lineNo++,HOURS,MINUTES,SECONDS,temperature, voltage[0],voltage[1],voltage[2],voltage[3], voltage[4],voltage[5],voltage[6],voltage[7]);
								
#ifdef UART_LOGGING_ENABLED
			printf("%s",(char*)dataString);
#endif
			
#ifdef SD_LOGGING_ENABLED
			for(int i=0; i<13; i++) fileNameTemp[i] = fileName[i];
			error = writeFile((uint8_t*)fileNameTemp);
			if(error) 
			{
				LCD_DisplayString_F(2, 1, PSTR(" SD Card Error  "));
				printf_P(PSTR("\n\n Error in writing to File.."));
				blinkRedLED();
			}
#endif			
		}
				
		do{
			_delay_ms(10);
			if((tick_1ms - time1) >= 999)
			{
				time1 = tick_1ms;
				updateDisplay();
			}
			if(SET_PRESSED) break;
			time3 = tick_1ms;
		}while(INTERVAL >= (time3 - time2));		//loop delay of INTERVAL ms (defined in main.h)
		time2 = time3;
	
		if(time3 >= MAX_TICK_VALUE)		//this condition comes only if datalogging is going on for more that 48 days		
		{
			 tick_1ms = 0;
			 time1 = time2 = time3 = 0;
		}
		
	} //end of while(1)
	
	return 0;
}

//******* Function to update date/time/temperature values on LCD **********
void updateDisplay(void)
{
	RTC_read();	
	sprintf(line1,"%02x/%02x  %02x:%02x:%02x ", DATE, MONTH, HOURS, MINUTES, SECONDS);
	temperature = getTemperature();
	if(screen == 0)
	{
		if(loggingON) sprintf(line2,"T:%+2.1f  Log:ON  ", temperature);
		else sprintf(line2,"T:%+2.1f  Log:OFF ", temperature);
	}
	
	LCD_DisplayString (1, 1, line1);
	LCD_DisplayString (2, 1, line2);
}

//***** Function to respond to any button-press during RTC set*******
void buttonResponseRTC(void)
{
	if(SET_PRESSED)
	{
		switch(screen)
		{
			case 0: sprintf(line2, "Date: %02x        ", DATE);
					screen = 1;	
					break;
			case 1: sprintf(line2, "Month: %02x       ", MONTH);
					screen = 2;
					break;
			case 2: sprintf(line2, "Year: 20%02x      ", YEAR);
					screen = 3;
					break;
			case 3: sprintf(line2, "Hour: %02x        ", HOURS);
					screen = 4;
					break;
			case 4: sprintf(line2, "Minute: %02x      ", MINUTES);
					screen = 5;
					break;
			case 5: SECONDS = 0;
					screen = 6;
					break;
			default:break;
		}
	}
	
	if(UP_PRESSED)
	{
		switch(screen)
		{
			case 1: DATE++;
					if((DATE % 16) == 0x0a) DATE += 0x06;
					if(DATE >= 0x32) DATE = 0x01;
					sprintf(line2, "Date: %02x        ", DATE);
					break;
			case 2: MONTH++;
					if(MONTH == 0x0a) MONTH = 0x10;
					if(MONTH >= 0x13) MONTH = 0x01;
					sprintf(line2, "Month: %02x       ", MONTH);
					break;
			case 3: YEAR++;
					if((YEAR % 16) == 0x0a) YEAR += 0x06;
					if(YEAR >= 0xa0) YEAR = 0x01;
					sprintf(line2, "Year: 20%02x      ", YEAR);
					break;
			case 4: HOURS++;
					if(HOURS == 0x0a) HOURS = 0x10;
					if(HOURS == 0x1a) HOURS = 0x20;
					if(HOURS >= 0x24) HOURS = 0x00;
					sprintf(line2, "Hour: %02x        ", HOURS);
					break;
			case 5: MINUTES++;
					if((MINUTES % 16) == 0x0a) MINUTES += 0x06;
					if(MINUTES >= 0x60) MINUTES = 0x00;
					sprintf(line2, "Minute: %02x      ", MINUTES);
					break;
			default:break;
		}
	}
	
	if(DWN_PRESSED)
	{
		switch(screen)
		{
			case 1: DATE--;
					if(DATE == 0x00) DATE = 0x31;
					else if((DATE % 16) == 0x0f) DATE -= 0x06;
					sprintf(line2, "Date: %02x        ", DATE);
					break;
			case 2: MONTH--;
					if(MONTH == 0x00) MONTH = 0x12;
					else if(MONTH == 0x0f) MONTH = 0x09;
					sprintf(line2, "Month: %02x       ", MONTH);
					break;
			case 3: if(YEAR == 0x00) YEAR = 0x99;
					else YEAR--;
					if((YEAR % 16) == 0x0f) YEAR -= 0x06;
					sprintf(line2, "Year: 20%02x      ", YEAR);
					break;
			case 4: if(HOURS == 0) HOURS = 0x23;
					else HOURS--;
					if(HOURS == 0x1f) HOURS = 0x19;
					if(HOURS == 0x0f) HOURS = 0x09;
					sprintf(line2, "Hour: %02x        ", HOURS);
					break;
			case 5: if(MINUTES == 0) MINUTES = 0x59;
					else MINUTES --;
					if((MINUTES % 16) == 0x0f) MINUTES -= 0x06;
					sprintf(line2, "Minute: %02x      ", MINUTES);
					break;
			default:break;
		}
	}
}

//function to blink LED in case of any error
void blinkRedLED(void)
{
  while(1)  //blink red LED continuously, if error
  {
    RED_LED_ON;
	_delay_ms(400);
	RED_LED_OFF;
	_delay_ms(400);
  }
}
  
//************************ END ****************************
